;; CLIPS Wine Advisor,
;; after an OPS5 program by Rand Waltzman of Teknowledge
;; Choosing a wine to go with a meal,
;; based on properties of the dish and tastes of the diner

;; TEMPLATES

(deftemplate wine
 (field property (type SYMBOL))
 (field is (type SYMBOL))
 (field cert (type FLOAT))
)

(deftemplate meal 
 (field property (type SYMBOL))
 (field is (type SYMBOL))
)

(deftemplate decision
 (field re (type SYMBOL))
 (field is (type SYMBOL))
)

;; FACTS

(deffacts the-facts
  (task dish)
)
; Production rules of the form

; (defrule <rulename> <RHS> => <LHS>).

;; RULES THAT FIND OUT ABOUT THE DISH.

; This corresponds to the data abstraction stage.

; dish-type is the rule invoked first because its calling
; pattern is put into WM by the fact list when CLIPS is reset.

(defrule dish-type
   (initial-fact)
   (task dish)
    => 
   (printout
	t crlf
	"Is the main dish of the meal MEAT, FISH or POULTRY?"
	t crlf)
   (assert (meal (property dish-type) (is (read))))
)

; meat finds out more about the dish, if it is meat.

(defrule meat
   (task dish)
   (meal (property dish-type) (is meat))
    => 
   (printout
	t crlf	
	"What kind of meat? e.g. STEAK, VEAL, LAMB"
	t crlf)
   (assert (meal (property meat-type) (is (read))))
)

; You can write a similar rule for fish distinguishing, e.g.
; wet fish from shell fish; also a rule for poultry.

;; RULES THAT SUGGEST WINE PROPERTIES

; This is the heuristic matching stage.

; steak is an example of a meat rule which suggests wine properties.

(defrule steak
    ?task <- (task dish)
    (meal (property meat-type) (is steak))
    => 
    (assert (wine (property colour) (is red) (cert 1.0)))
    (assert (wine (property body) (is full) (cert 1.0)))
    (assert (wine (property flavour) (is dry) (cert 0.7)))
    (assert (wine (property flavour) (is sweet) (cert 0.2)))
    (retract ?task)
    (assert (task attributes))
)

; You can write similar rules for lamb, veal, etc.
; We remove the dish task once we know what the dish is. 
; Then we set up the task of deciding upon wine attributes.

; You should have default rules for meat, fish and poultry, which
; catch any cases, such as alligator, that you have not anticipated.

;; RULES THAT HANDLE CERTAINTIES

; If there are two structures in WM with the same value for
; the same attribute but with different certainties
; then attribute-update creates a third structure with a new value
; and deletes the other two.

; The formula for computing the new CF is
; 		cf = cf1 + cf2(1 - cf1)

(defrule attribute-update
    (task attributes)
    ?wine1 <- (wine (property ?attribute) (is ?value) (cert ?cert1))
    ?wine2 <- (wine (property ?attribute) (is ?value) (cert ?cert2))
    (<> ?cert1 ?cert2)
    => 
    (bind ?newcert (+ ?cert1 (* ?cert2 (- 1 ?cert1))))
    (assert (wine (property ?attribute) (is ?value) (cert ?newcert))) 
    (retract ?wine1)
    (retract ?wine2)
)

; Write a rule called preference, which is only invoked if there is 
; more than one possible value for an attribute in working memory.
; The rule should ask the user's preference and change the certainty 
; of the relevant attribute-value combination to unity, deleting the
; other values from working memory.

; Write a rule called choose-value which fires if there are two
; structures in working memory which carry different values
; under the same attribute.  The rule should choose the structure
; with the greater CF and delete the other.

; Write a rule called unique, which fires if there is only one
; candidate value for an attribute, and declares the attribute done.

; Write a rule called unity, which fires if a particular value is
; definite (has CF = 1) and declares the attribute done.

; if all the attributes of the wine are done then report them

(defrule all-attributes-done
    ?task <- (task attributes)
    ?col <- (colour done)
    ?bod <- (body done)
    ?fla <- (flavour done)
    (wine (property colour) (is ?colour))
    (wine (property body) (is ?body))
    (wine (property flavour) (is ?flavour))
    => 
    (printout
	t crlf
	"Try a " ?flavour " " ?colour " wine with a " ?body " body"
	t crlf)
    (retract ?col)
    (retract ?bod)
    (retract ?fla)
    (retract ?task)
    (assert (task brand))
)

; RULES FOR THE WINE

; Select a wine, given an attribute-value description.
; This corresponds to the solution refinement stage.
; Generate all the candidates before you offer selections to the user.

; Write a rule called go-choose which selects the highest scoring
; wine when no other rules will fire to propose any more wines.

; soave is a sample wine rule.  Write as many as you like.

(defrule soave
    (task brand)
    (wine (property colour) (is white))
    (wine (property flavour) (is dry) (cert ?cert1))
    (wine (property body) (is fine) (cert ?cert2))
    => 
    (assert
	(wine (property brand) (is soave) (cert (min ?cert1 ?cert2))))
)

; RULES FOR PRESENTING SELECTIONS

; The user responds to these by typing 'yes' or 'no'.

; Write a rule called selection which finds the highest scoring wine
; and offers it to the user.

; Write a rule called rejection which deals with a negative response
; to the current suggestion.

; acceptance ends the session on the correct obsequious note.

(defrule acceptance
    (task choice)
    (decision (re ?candidate) (is yes))
    => 
    (printout t crlf "Sir/Madam has impeccable taste" t crlf)
    (halt)
)



